;********************************************************************
;*                                                                  *
;* Title       : DCF                                                *
;* Version     : 1.3_629    06. Oktober 2006                        *
;* Author      : Uwe Nagel, Parkstr. 46, D-68766 Hockenheim         *
;* Assembler   : MPASM (c) Microchip                                *
;*                                                                  *
;********************************************************************
; 0.1    21.05.02 erster Entwurf
; 0.2    30.05.02 erste lauffhige Version mit LCD
; 0.3    29.09.02 Kosmetik am Quellcode und an der Homepage, keine
;                 nderung der Funktion
; 0.4    01.04.03 Antennen Symbol auf Wunsch von Christof Rue
; 0.5    24.07.03 Zeitimpule auf Wunsch von CIOVALLADOLID <CIOVALLADOLID@terra.es>
;                 wahlweise zwei feste Schaltausgnge auf Wunsch von Herrn Eichler
; 0.6    30.08.03 Ports lschen beim Einschalten
;                 PTT-Signal auf PORTA,4 (CIOVALLADOLID)
; 0.7    26.09.03 Filter gegen unsichere Flanken
; 0.8    01.10.03 diverse Optimierungen
; 0.9    15.08.04 anderes Ausgabeformat auf Wunsch von Peter Hilkmann (PE1DCD)
; 0.9a   02.11.04 Fehler bei Monatsausgabe beseitigt, neue Sprachen
; 0.a    06.12.04 Spezialversion mit 16F874 und 4MHz (hier nicht integriert)
; 0.b    03.04.05 serielle Ausgabe der Zeit, Sprachen ausgelagert
; 0.ba   23.10.05 neue Sprache 'swedish', Dank an Volker Lange-Janson
; 0.bb   09.02.06 fehlendes 'RETURN' eingefgt, Dank an Peter Engels
; 0.bc   10.02.06 Antennensymbol auch bei PE1DCD ermglichen
;                 Programm adaptiert auf 16F628 (Nachfolger 16F84)
; 0.bd   21.02.06 useTimer1 eingefhrt
; 0.c    18.03.06 erster Versuch mit Nebenuhr
; 1.0    08.04.06 Hauptuhr mit Sommer-Winterzeit-Umstellung
; 1.1    19.06.06 andere Impulsausgabe fr Mutteruhr (PULS500)
; 1.2    02.09.06 blinkendes Antennensymbol
; 1.3    28.09.06 Plausibilittskontrolle von Puls- und Pausenzeiten
; 1.3_629 06.10.06 Sparversion ohne LCD als Mutteruhr
;********************************************************************


#define TITEL "-DCF77-Uhr 1.3_629-"

;********************************************************************
; Pin 8 GND
; Pin 7 GPIO.0 DCF Eingang
; Pin 6 GPIO.1 LED blinkt im Sekundentakt
; Pin 5 GPIO.2 Ausgang an Verstrker
; Pin 4 GPIO.3 auf GND ziehen zum Stellen
; Pin 3 GPIO.4 Quarz 3,2768MHz
; Pin 2 GPIO.5 Quarz 3,2768MHz
; Pin 1 VCC
;********************************************************************
; nicht synchronisiert: Led leuchtet, wenn DCF-Eingang aktiv ist
; synchronisiert: Led leuchtet, wenn DCF-Eingang _nicht_ aktiv ist
;********************************************************************

;--- MUTTERUHR definieren um auf GPIO.2 einen Minutentakt zum Ansteuern 
; von Nebenuhren auszugeben. GPIO.3 muss dazu auf 1 liegen. Zieht man GPIO.3
;  auf Masse, dann wird jede Sekunde ein Takt zum Stellen ausgegeben.
#define MUTTERUHR
; wenn PULS500 definiert ist, wird jede Minute ein 500ms-Puls ausgegeben
; sonst wechselt GPIO,2 jede Minute den Pegel
;#define PULS500


	title		TITEL

	__config	_XT_OSC & _WDT_ON & _PWRTE_ON & _CP_OFF & _MCLRE_OFF

	errorlevel	-302		; keine bank select message
	radix	dec

	include	<P12F629.INC>

;****** Variablen ***************************************************
 cblock	0x20
	save_W		; w und
	save_STATUS	; STATUS whrend interrupt
	save_FSR	; FSR whrend Interrupt
	flag		; diverse Bits
	temp
	tmp1
	tmp2
	zweihund	; 200 Interupts sind 1 Sek
	puls_zeit	; misst Pulsdauer
	pause_zeit	; misst Pausendauer
;*** diese Zeit wird angezeigt
	dispSec, dispMin, dispStd, dispTag, dispMon, dispJahr, dispWtag, dispStat
;*** hier kommt die gerade empfangene Zeit rein
	dcfMin, dcfStd, dcfTag, dcfMon, dcfJahr, dcfWtag, dcfStat
;*** zum Vergleich, die letzte Minute
	dcf1Min, dcf1Std, dcf1Tag, dcf1Mon, dcf1Jahr, dcf1Wtag, dcf1Stat
	filter, altsec, flag2

	pulshi, pulslo, pulsdauer
 endc


;****** bits in flag ************************************************
#define	neusync flag,0
#define	alt     flag,1
#define	sekunde	flag,2
#define	syncd	flag,3
#define	dcfsign	flag,4
#define ms5	flag,5
#define	comm	flag,7

#define SommerWinter	flag2,0
#define	WinterSommer	flag2,1
#define Pulsstart	flag2,2
#define Pulsende	flag2,3



;****** Konstanten **************************************************
; TICKS berechnet sich folgendermassen (fr Timer0):
; Oszillatorfrequenz/4/256/Prescaler/TICKS=1 Hz
; oder
; TICKS=Oszillatorfrequenz/4/256/Prescaler
; wenn Ticks nicht ganzzahlig ist, 
; geht die Uhr im nicht synchronisierten Betrieb ungenau
; bei useTimer1 und 4MHz muss auch 200 stehen
#define	TICKS	200	; fr 3,2768 MHz Quarz bei 16F84 4MHz bei 16F628


;********************************************************************
;* Programmcode ...
;********************************************************************
	org	0
	goto	main

;********************************************************************
;* interrupt :
;********************************************************************
	org	4
	goto	interrupt
;********************************************************************
;* Tabellen in Page 0:
;********************************************************************



;********************************************************************
;* interrupt  service Routine,
;* hinter Tabellen, damit die alle in Page 0 liegen knnen:
;********************************************************************
interrupt
	movwf	save_W		; W und STATUS sichern
	swapf	STATUS,w
	movwf	save_STATUS
	movf	FSR,w
	movwf	save_FSR

	bcf	STATUS,RP0	; Whle register page 0

	btfss	INTCON,T0IF
	goto	no_timer_isr ; ist es timer interrupt ?


#ifdef PULS500
	movf	pulsdauer,w
	btfsc	STATUS,Z
	bcf	GPIO,2
	btfss	STATUS,Z
	decf	pulsdauer,f
#endif


; increment filter if port=1 and filter if less than 8
; decrement filter if port=0 and filter if greater than 0


	btfss	GPIO,0

	goto	port_is_0
port_is_1
	incf	filter,f
	btfsc	filter,3	; <8 ?
port_is_0
	decf	filter,f
	btfsc	filter,7
	clrf	filter

; set dcfsign if filter becomes 6
; clear dcfsign if filter becomes 1
; 0 1 2 3 4 5 6 7   6 5 4 3 2 1 0
; 0 0 0 0 0 0 1 1   1 1 1 1 1 0 0

	movf	filter,w
	xorlw	6
	btfsc	STATUS,Z
	bsf	dcfsign
	xorlw	6^1
	btfsc	STATUS,Z
	bcf	dcfsign

;** teste auf steigende Flanke
	btfss	alt
	btfss	dcfsign
	goto	NoPos

;** Pause krzer als 750ms ist ungltig	
	movf	pause_zeit,w
	addlw	-150		; >150*5ms=750ms
	btfss	STATUS,C
	goto	NoPos

;** positive Flanke...
	bsf	alt
	bsf	Pulsstart
	clrf	puls_zeit

	btfss	neusync	; erster Impuls nach 59. Sekunde?
	goto	NoNeg

	clrf	dispSec
	movf	dcfMin,w
	movwf	dispMin
	movf	dcfStd,w
	movwf	dispStd
	movf	dcfTag,w
	movwf	dispTag
	movf	dcfWtag,w
	movwf	dispWtag
	movf	dcfMon,w
	movwf	dispMon
	movf	dcfJahr,w
	movwf	dispJahr
	movf	dcfStat,w
	movwf	dispStat

	bcf	neusync
	movlw	TICKS	; 200 * 5ms
	movwf	zweihund
	goto	end_timer_isr

NoPos:
;** teste auf fallende Flanke
	btfsc	alt
	btfsc	dcfsign
	goto	NoNeg

;** Puls krzer als 80ms ist ungltig	
	movf	puls_zeit,w
	addlw	-16		; >16*5ms=80ms
	btfss	STATUS,C
	goto	NoNeg

;** fallende Flanke
	bcf	alt
	clrf	pause_zeit
	bsf	Pulsende

	movf	puls_zeit,w
	addlw	-30		; >30*5ms
	call	shiftDCF

NoNeg:
	incfsz	pause_zeit,w	; bis max 255 zhlen damit
	incf	pause_zeit,f	; bei lngeren Pausen nicht wieder
				; bei 0 angefangen wird

	decfsz	zweihund,f
	goto	TimeOut

;** Sekunde abgelaufen
	movlw	TICKS
	movwf	zweihund
	call	tick

TimeOut:
	incfsz	puls_zeit,f	; 1,28 s keine steigende Flanke ?
	goto	end_timer_isr

	call	sec59		; berprfe empfangene Zeit

end_timer_isr
	bcf	sekunde
	movf	dispSec,w
	xorwf	altsec,w	; hat sich die Sekunde gendert ?
	skpz
	bsf	sekunde
	xorwf	altsec,f	; aktuelle Sekunde merken

	bcf	INTCON,T0IF     ; lsche timer interrupt flag

no_timer_isr
	movf	save_FSR,w
	movwf	FSR
	swapf	save_STATUS,w
	movwf	STATUS
	swapf	save_W,f
	swapf	save_W,w
	retfie






;********************************************************************
;* Hauptprogramm :
;********************************************************************
main    
	movlw 	0x07
	movwf	CMCON 	; disable comparators (= GPIO's sind digital)

	bsf	STATUS,RP0	; select Register-Page 1
	movlw	B'00000011'	; TMR0 internal clock/16 = 200Hz @ 3,2768MHz Quarz
	movwf	OPTION_REG

	movlw	B'00111001'
	movwf	TRISIO

#ifdef	__12F675
	clrf	ANSEL		; beim 12F675 muss auch der AD_Wandler abgeschaltet werden
#endif

	bcf	STATUS,RP0	; select Register-Page 0

	clrf	GPIO
	clrf	PCLATH

	clrf	flag
	clrf	puls_zeit
	movlw	TICKS
	movwf	zweihund

	clrf	dispSec
	clrf	dispMin
	clrf	dispStd
	clrf	dispWtag
	clrf	dispJahr
	movlw	1
	movwf	dispMon
	movwf	dispTag

	clrf	flag2
	clrf	pulslo
	clrf	pulshi

	movlw	B'10100000'	; setze GIE und T0IE
	movwf	INTCON	

mainloop:
	clrwdt	

	btfsc	syncd
	goto	syncLED
; nicht synchronisiert, Led leuchtet, wenn Puls aktiv ist
	btfsc	alt
	bsf	GPIO,1
	btfss	alt
	bcf	GPIO,1
	goto	goon
syncLED
; synchronisiert, Led leuchtet, wenn Puls _nicht_ aktiv ist
	btfsc	alt
	bcf	GPIO,1
	btfss	alt
	bsf	GPIO,1
goon


	btfss	sekunde
	goto	mainloop
	bcf	sekunde



	call	special

	goto	mainloop



;********************************************************************
;* Spezialprogramme fr verschiedene Zwecke
;********************************************************************
special
	btfss	WinterSommer
	goto	Mutter1
	bcf	WinterSommer
	movlw	low 62
	movwf	pulslo
	movlw	high 62
	movwf	pulshi
Mutter1
	btfss	SommerWinter
	goto	Mutter2
	bcf	SommerWinter
	movlw	low 672
	movwf	pulslo
	movlw	high 672
	movwf	pulshi
Mutter2
	movf	pulslo,w
	iorwf	pulshi,w
	skpnz
	goto	Mutter3

	decf	pulslo,f
	incf	pulslo,w
	skpnz
	decf	pulshi,f

	goto	stellen

Mutter3

	btfss	GPIO,3
	goto	stellen

	movf	dispSec,w
	skpz
	return

stellen
	movlw	0x04
	xorwf	GPIO,f
#ifdef PULS500
	movlw	100
	movwf	pulsdauer
#endif
	return



;********************************************************************

shiftDCF:
	rrf	dcfJahr,f
	rrf	dcfMon,f
	rrf	dcfTag,f
	rrf	dcfStd,f
	rrf	dcfMin,f
	rrf	dcfStat,f
	return

;********************************************************************
;* in der 59. Sekunde aufgerufen, sortiert erst mal den Empfangspuffer,
;* um die Zeitwerte in einzelnen Bytes zu erhalten
;********************************************************************
;55555555 54444444 44433333 33333222 22222221 11111111
;87654321 09876543 21098765 43210987 65432109 87654321   Sekunde
;PJJJJJJJ JMMMMMWW WTTTTTTP SSSSSSPm mmmmmm1a ooaR0000   DCF-Bits
;
;JJJJJJJJ 000MMMMM 00tttttt 00ssssss 0mmmmmmm 1aooaR00 00000www
; Jahr     Monat    Tag      Stunde   Minute   Status   Wtag
;********************************************************************

sec59:	rlf	dcfMon,w
	rlf	dcfJahr,f	; =Jahr
	rlf	dcfTag,w
	rlf	dcfMon,w
	andlw	0x07
	movwf	dcfWtag		; =Wochentag Mo..So = 1..7
	rrf	dcfMon,f
	rrf	dcfMon,f
	movlw	0x1f
	andwf	dcfMon,f	; =Monat
	rrf	dcfTag,f
	movlw	0x3f
	andwf	dcfTag,f	; =Tag
	rrf	dcfStd,w
	rrf	dcfMin,f
	rrf	dcfStat,f
	rrf	dcfMin,f
	rrf	dcfStat,f	; =Status
	movlw	0x7f
	andwf	dcfMin,f	; =Minute
	rrf	dcfStd,f
	rrf	dcfStd,f
	movlw	0x3f
	andwf	dcfStd,f	; =Stunde

	movlw	dcf1Min
	movwf	FSR
	call	tickMin		; letzte Zeit weiterrechnen

	movf	dcfMin,w	; und mit gerade empfangener
	xorwf	dcf1Min,w	; Zeit vergleichen
	skpz
	goto	ungleich
	movf	dcfStd,w
	xorwf	dcf1Std,w
	skpz
	goto	ungleich
	movf	dcfTag,w
	xorwf	dcf1Tag,w
	skpz
	goto	ungleich
	movf	dcfMon,w
	xorwf	dcf1Mon,w
	skpz
	goto	ungleich
	movf	dcfJahr,w
	xorwf	dcf1Jahr,w
	skpz
	goto	ungleich

;*** zwei aufeinanderfolgende Zeiten haben zueinander gepasst
;*** die aktuelle Zeit bei nchster Sekunde ins Display kopieren

	bsf	syncd
	bsf	neusync
	return

ungleich:
;*** zwei aufeinanderfolgende Zeiten haben nicht eine Minute unterschied
;*** merken wir die gerade empfangene Zeit fr die nchste Minute
	movf	dcfMin,w
	movwf	dcf1Min
	movf	dcfStd,w
	movwf	dcf1Std
	movf	dcfTag,w
	movwf	dcf1Tag
	movf	dcfMon,w
	movwf	dcf1Mon
	movf	dcfJahr,w
	movwf	dcf1Jahr
	movf	dcfStat,w
	movwf	dcf1Stat
	bcf	syncd
	return

;********************************************************************
;* hier wird eine Zeit um eine Minute weitergezhlt.
;********************************************************************
tick:	movlw	dispSec
	movwf	FSR

	call	IncDec	; Sekunde
	movlw	0x60
	xorwf	INDF,w
	skpz
	return
	clrf	INDF	; Sekunde=0

	incf	FSR,f	; ->Minute
tickMin:
	call	IncDec	; Minute
	movlw	0x60
	xorwf	INDF,w
	skpz
	return
	clrf	INDF	; Minute=0
	incf	FSR,f	; -> Stunde
	call	IncDec	; Stunde

	movf	FSR,w	; Zeiger auf Stunde sichern
	movwf	tmp2

	movlw	4
	addwf	FSR,f	; -> WTag
	movf	INDF,w
	xorlw	7	; Sonntag?
	skpz
	goto	keinWechsel

; es ist ein Sonntag
	decf	FSR,f	; ->Jahr
	decf	FSR,f	; ->Monat
	decf	FSR,f	; ->Tag
	movf	INDF,w
	addlw	-0x25
	btfss	STATUS,C
	goto	keinWechsel	; Tag ist kleiner 25

; es ist mindestens der 25.
	incf	FSR,f	; -> Monat
	movf	INDF,w
	xorlw	0x03	; Mrz?
	skpz
	goto	keinMaerz	; kein Mrz

; es ist ein Sonntag >=25. im Mrz
	decf	FSR,f	; -> Tag
	decf	FSR,f	; -> Stunde
	movf	INDF,w
	xorlw	0x02
	skpz
	goto	keinWechsel	; es ist nicht 2Uhr

;es ist 2 Uhr --> es wird Sommer
	incf	INDF,f	; Uhr eine Stunde vorstellen
	bsf	WinterSommer
	movlw	5
	addwf	FSR,f	; -> Status
	movlw	0x30
	xorwf	INDF,f	; Status umstellen
	goto	fertig

keinMaerz:
	xorlw	0x03^0x10	; Oktober?
	skpz
	goto	keinWechsel

; es ist ein Sonntag >=25. im Oktober
	decf	FSR,f	; -> Tag
	decf	FSR,f	; -> Stunde
	movf	INDF,w
	xorlw	0x03
	skpz
	goto	keinWechsel	; es ist nicht 3 Uhr

; es ist ein Sonntag >=25. im Oktober, 3 Uhr
	movlw	5
	addwf	FSR,f	; -> Status
	movf	INDF,w
	andlw	0x10	; ist noch Sommerzeit?
	skpnz
	goto	keinWechsel

	movlw	0x30
	xorwf	INDF,f	; Status umstellen
	movf	tmp2,w
	movwf	FSR	; -> Stunde
	decf	INDF,f	; Uhr eine Stunde zurck
	bsf	SommerWinter
	goto	fertig

keinWechsel:
	bcf	SommerWinter
	bcf	WinterSommer
fertig:
	movf	tmp2,w
	movwf	FSR		; Zeiger wieder auf Stunde

	movlw	0x24	; Stunde==24?
	xorwf	INDF,w
	skpz
	return
	clrf	INDF	; Stunde=0

	incf	FSR,f
tickTag:
	call	IncDec	; Tag
	
	incf	FSR,f	; ->Monat
	incf	FSR,f	; ->Jahr
	incf	FSR,f	; ->WTag
	incf	INDF,f
	movlw	1
	btfsc	INDF,3	; WTag==8?
	movwf	INDF
	decf	FSR,f	; ->Jahr
	decf	FSR,f	; ->Monat

	movlw	8
	subwf	INDF,w
	movf	INDF,w
	skpnc
	addlw	-7
	andlw	0x01	
	addlw	0x31
	movwf	temp
	movf	INDF,w
	xorlw	2	; Februar ?
	skpz
	goto	nofeb
	movlw	0x30
	movwf	temp
	incf	FSR,f	; ->Jahr
	movf	INDF,w	; Jahr
	decf	FSR,f	; ->Monat
	decf	FSR,f	; -> Tag

; 00 04 08 12 16 sind Schaltjahre, danach wiederholt es sich
	andlw	0x13
	skpnz
	goto	leapyear

	xorlw	0x12
	skpnz
	goto	leapyear
	movlw	0x29
	movwf	temp
leapyear:
nofeb:	
	movf	temp,w
	xorwf	INDF,w
	skpz
	return
	movlw	1
	movwf	INDF	; Tag=1

	incf	FSR,f	; -> Monat
	call	IncDec	
	movf	INDF,w
	xorlw	0x13
	skpz
	return
	movlw	1
	movwf	INDF	; Monat=1

	incf	FSR,f	; -> Jahr
;	call	IncDec
;	return

IncDec:	incf	INDF,f		; Zhler erhhen
	movlw	6
	addwf  	INDF,w		; +6 (Dez.-korrektur) --> Ergebnis vorerst in w
	btfsc	STATUS,DC	; berlauf im Low-Nibble?
	movwf	INDF		; ja, Korr. Ergebn. zurck
	return

;********************************************************************
	end	
